from typing import Iterable

from fdsreader.utils import Mesh
from fdsreader.utils.data import FDSDataCollection


class MeshCollection(FDSDataCollection):
    """Collection of :class:`Obstruction` objects. Offers extensive functionality for filtering and
        using obstructions as well as dependent such as :class:`Boundary`.
    """

    def __init__(self, *meshes: Iterable[Mesh]):
        super().__init__(*meshes)

    def get_by_id(self, mesh_id: str):
        """Get the mesh with corresponding id if it exists.
        """
        return next((mesh for mesh in self if mesh.id == mesh_id), None)

    def __repr__(self):
        return "MeshCollection(" + super(MeshCollection, self).__repr__() + ")"



from typing import List, Tuple, Dict, Sequence, Union

import numpy as np

from fdsreader.utils import Quantity, Mesh, Extent


# class Entrance:
#     def __init__(self):
#         self.id = ""
#         self.extent = Extent(1.0, 1.0, 1.0, 1.0, 1.0, 1.0)
#         self.orientation = 1
#         self.color = (0.0, 0.0, 0.0)


# class Exit:
#     def __init__(self):
#         self.id = ""
#         self.extent = Extent(1.0, 1.0, 1.0, 1.0, 1.0, 1.0)
#         self.xyz = (1.0, 1.0, 1.0)
#         self.orientation = 1
#         self.color = (0.0, 0.0, 0.0)
#         self.count_only = True
#         self.known_door = True


# class Door:
#     def __init__(self):
#         self.id = ""
#         self.extent = Extent(1.0, 1.0, 1.0, 1.0, 1.0, 1.0)
#         self.xyz = (1.0, 1.0, 1.0)
#         self.orientation = 1
#         self.color = (0.0, 0.0, 0.0)
#         self.to_node = Entrance()
#         self.known_door = True
#         self.exit_sign = True


class Evacuation:
    """Container to store evac data from evac simulations with FDS.

    :ivar class_name: Name of the evac class defined in the FDS input-file.
    :ivar quantities: List of all quantities for which data has been written out.
    :ivar color: Color assigned to the evac.
    :ivar n_humans: Number of existing evacs for each timestep per mesh.
    :ivar lower_bounds: Dictionary with lower bounds for each timestep with quantities as keys.
    :ivar upper_bounds: Dictionary with upper bounds for each timestep with quantities as keys.
    """

    def __init__(self, class_name: str, quantities: List[Quantity], color: Tuple[float, float, float]):
        self.class_name = class_name
        self.quantities = quantities
        self.color = color
        self.n_humans: Dict[Mesh, List[int]] = dict()

        self._positions: List[np.ndarray] = list()
        self._body_angles: List[np.ndarray] = list()
        self._semi_major_axis: List[np.ndarray] = list()
        self._semi_minor_axis: List[np.ndarray] = list()
        self._agent_heights: List[np.ndarray] = list()
        self._tags: List[np.ndarray] = list()
        self._data: Dict[str, List[np.ndarray]] = {q.name: [] for q in self.quantities}
        self.times: Sequence[float] = list()

        self.lower_bounds = {q.name: [] for q in self.quantities}
        self.upper_bounds = {q.name: [] for q in self.quantities}

        self._init_callback = None

    @property
    def id(self):
        return self.class_name

    def has_quantity(self, quantity: Union[Quantity, str]):
        if type(quantity) == Quantity:
            quantity = quantity.name
        return any(
            q.name.lower() == quantity.lower() or q.short_name.lower() == quantity.lower() for q in self.quantities)

    def filter_by_tag(self, tag: int):
        """Filter all evacs by a single one with the specified tag.
        """
        data = self.data
        tags = self.tags
        positions = self.positions
        evac = Evacuation(self.class_name, self.quantities, self.color)
        evac._tags = tag

        evac._data = {quantity: list() for quantity in data.keys()}
        evac._positions = list()
        evac.times = list()

        for t, tags in enumerate(tags):
            if tag in tags:
                idx = np.where(tags == tag)[0]

                for quantity in data.keys():
                    evac._data[quantity].append(data[quantity][t][idx][0])
                evac._positions.append(positions[t][idx][0])
                evac.times.append(self.times[t])

        evac.lower_bounds = dict()
        evac.upper_bounds = dict()
        for q in self.quantities:
            if len(evac._positions) != 0:
                evac.lower_bounds[q.name] = np.min(evac._data[q.name])
                evac.upper_bounds[q.name] = np.max(evac._data[q.name])
            else:
                evac.lower_bounds[q.name] = 0
                evac.upper_bounds[q.name] = 0

        return evac

    @property
    def data(self) -> Dict[str, List[np.ndarray]]:
        """Dictionary with quantities as keys and a list with a numpy array for each timestep which
            contains data for each person in that timestep.
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._data

    def get_data(self, quantity: Union[Quantity, str]) -> List[np.ndarray]:
        """Returns a list with a numpy array for each timestep which contains data about the specified quantity for
            each person in that timestep.
        """
        if self.has_quantity(quantity):
            if type(quantity) == Quantity:
                quantity = quantity.name
            return self.data[quantity]
        return []

    @property
    def tags(self) -> List[np.ndarray]:
        """List with a numpy array for each timestep which contains a tag for each evac in that
            timestep.
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._tags

    @property
    def positions(self) -> List[np.ndarray]:
        """List with a numpy array for each timestep which contains the position of each evac in
            that timestep.
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._positions

    @property
    def body_angles(self) -> List[np.ndarray]:
        """
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._body_angles

    @property
    def semi_major_axis(self) -> List[np.ndarray]:
        """
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._semi_major_axis

    @property
    def semi_minor_axis(self) -> List[np.ndarray]:
        """
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._semi_minor_axis

    @property
    def agent_heights(self) -> List[np.ndarray]:
        """
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._agent_heights

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if len(self._positions) != 0:
            del self._positions
            self._positions = list()
            del self._tags
            self._tags = list()
            del self._data
            self._data = {q.name: [] for q in self.quantities}

    def __repr__(self):
        return f"Evacuation(name={self.class_name}, quantities={self.quantities})"



import os
from copy import deepcopy

import numpy as np
import logging
from typing import Dict, Collection, Tuple, Union, List
from typing_extensions import Literal

from fdsreader.utils import Dimension, Quantity, Mesh, Extent
from fdsreader import settings
import fdsreader.utils.fortran_data as fdtype

_HANDLED_FUNCTIONS = {}


def implements(np_function):
    """Decorator to register an __array_function__ implementation for GeomSlices.
    """

    def decorator(func):
        _HANDLED_FUNCTIONS[np_function] = func
        return func

    return decorator


class SubGeomSlice:
    """Part of a geomslice that cuts through a single mesh.

    :ivar mesh: The mesh the subgeomslice cuts through.
    """

    def __init__(self, parent_slice, filename: str, geom_filename: str, extent: Extent, mesh: Mesh):
        self._parent_slice = parent_slice
        self.mesh = mesh
        self.extent = extent

        self.file_path = os.path.join(parent_slice._root_path, filename)
        self.geom_file_path = os.path.join(parent_slice._root_path, geom_filename)

        # self.vector_filenames = dict()
        # self._vector_data = dict()

    @property
    def orientation(self) -> Literal[1, 2, 3]:
        """Orientation [1,2,3] of the geomslice in case it is 2D, 0 otherwise.
        """
        return self._parent_slice.orientation

    @property
    def times(self):
        return self._parent_slice.times

    @property
    def n_t(self):
        return self._parent_slice.n_t

    def _load_geom_data(self):
        with open(self.geom_file_path, 'rb') as infile:
            dtype_meta = fdtype.new((('i', 3),))
            infile.seek(2 * fdtype.INT.itemsize + dtype_meta.itemsize + fdtype.FLOAT.itemsize)
            self.n_verts, self.n_faces, n_vols = fdtype.read(infile, dtype_meta, 1)[0][0]

            if self.n_verts > 0 and self.n_faces > 0:
                dtype_verts = fdtype.new((('f', 3 * self.n_verts),))
                dtype_faces = fdtype.new((('i', 3 * self.n_faces),))
                # dtype_locations = fdtype.new((('i', self.n_faces),))
                # dtype_zero_floats = fdtype.new((('f', 3 * self.n_faces * 2),))
                self._vertices = fdtype.read(infile, dtype_verts, 1)[0][0].reshape((self.n_verts, 3), order='F')
                self._faces = fdtype.read(infile, dtype_faces, 1)[0][0].reshape((self.n_faces, 3), order='F')
            else:
                self._vertices = np.array([])
                self._faces = np.array([])

    def _load_data(self):
        with open(self.file_path, 'rb') as infile:
            infile.seek(2 * fdtype.INT.itemsize)

            if self.n_verts > 0 and self.n_faces > 0:
                dtype_data = fdtype.combine(fdtype.FLOAT, fdtype.new((('i', 4),)), fdtype.new((('f', self.n_faces),)))
            else:
                dtype_data = fdtype.combine(fdtype.FLOAT, fdtype.new((('i', 4),)))

            load_times = self.n_t == -1
            if load_times:
                self._parent_slice.n_t = (os.stat(self.file_path).st_size - 2 * fdtype.INT.itemsize) // dtype_data.itemsize
                self._parent_slice.times = np.empty(self.n_t)

            self._data = np.empty((self.n_t, self.n_faces), dtype=np.float32)

            if self.n_verts > 0 and self.n_faces > 0:
                for t, data in enumerate(fdtype.read(infile, dtype_data, self.n_t)):
                    if load_times:
                        self.times[t] = data[0][0]
                    self._data[t, :] = data[2]

    @property
    def data(self) -> np.ndarray:
        """Method to lazy load the geomslice's data.
        """
        if not hasattr(self, "_data"):
            _ = self.vertices  # Make sure geom data has been loaded already
            self._load_data()
        return self._data

    @property
    def vertices(self) -> np.ndarray:
        """Method to lazy load the geomslice's data.
        """
        if not hasattr(self, "_vertices"):
            self._load_geom_data()
        return self._vertices

    @property
    def faces(self) -> np.ndarray:
        """Method to lazy load the geomslice's data.
        """
        if not hasattr(self, "_faces"):
            self._load_geom_data()
        return self._faces

    # @property
    # def vector_data(self) -> Dict[str, np.ndarray]:
    #     """Method to lazy load the geomslice's vector data if it exists.
    #     """
    #     if not hasattr(self, "_vector_data"):
    #         raise AttributeError("There is no vector data available for this geomslice.")
    #     if len(self._vector_data) == 0:
    #         for direction in self.vector_filenames.keys():
    #             file_path = os.path.join(self._parent_slice._root_path,
    #                                      self.vector_filenames[direction])
    #             self._vector_data[direction] = np.empty((self.n_t,) + self.shape,
    #                                                     dtype=np.float32)
    #             self._load_data(file_path, self._vector_data[direction])
    #     return self._vector_data

    @property
    def vmin(self):
        return np.min(self.data)

    @property
    def vmax(self):
        return np.max(self.data)

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if hasattr(self, "_data"):
            del self._data
            # del self._vector_data
            # self._vector_data = dict()

    def __repr__(self):
        return f"SubGeomSlice(mesh={self.mesh.id})"


class GeomSlice(np.lib.mixins.NDArrayOperatorsMixin):
    """Slice file data container including metadata. Consists of multiple subgeomslices, one for each
        mesh the geomslice cuts through.

    :ivar cell_centered: Indicates whether centered positioning for data is used.
    :ivar quantity: Quantity object containing information about the quantity calculated for this
        geomslice with the corresponding short_name and unit.
    :ivar times: Numpy array containing all times for which data has been recorded.
    :ivar n_t: Total number of time steps for which output data has been written.
    :ivar orientation: Orientation [1,2,3] of the geomslice in case it is 2D, 0 otherwise.
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    """

    def __init__(self, root_path: str, geomslice_id: str, times: np.ndarray, multimesh_data: Collection[Dict]):
        self._root_path = root_path

        self.times = times

        if times is not None:
            self.n_t = times.size
        else:
            self.n_t = -1

        self.id = geomslice_id

        # List of all subgeomslices this geomslice consists of (one per mesh).
        self._subgeomslices: Dict[Mesh, SubGeomSlice] = dict()

        vector_temp = dict()
        for mesh_data in multimesh_data:
            if "-VELOCITY" in mesh_data["quantity"]:
                vector_temp[mesh_data["mesh"]] = dict()

        for mesh_data in multimesh_data:
            if mesh_data["mesh"] not in self._subgeomslices:
                self.quantity = Quantity(mesh_data["quantity"], mesh_data["short_name"], mesh_data["unit"])
                self._subgeomslices[mesh_data["mesh"]] = SubGeomSlice(self, mesh_data["filename"], mesh_data["geomfilename"], mesh_data["extent"], mesh_data["mesh"])

            # if "-VELOCITY" in mesh_data["quantity"]:
            #     vector_temp[mesh_data["mesh"]][mesh_data["quantity"]] = mesh_data["filename"]

        # for mesh, vector_filenames in vector_temp.items():
        #     if "U-VELOCITY" in vector_filenames:
        #         self._subgeomslices[mesh].vector_filenames["u"] = vector_filenames["U-VELOCITY"]
        #     if "V-VELOCITY" in vector_filenames:
        #         self._subgeomslices[mesh].vector_filenames["v"] = vector_filenames["V-VELOCITY"]
        #     if "W-VELOCITY" in vector_filenames:
        #         self._subgeomslices[mesh].vector_filenames["w"] = vector_filenames["W-VELOCITY"]

        # # Iterate over all subgeomslices and remove duplicated ones which will be created when the geomslice
        # # cuts exactly through two mesh borders. Only required for non-cell-centered geomslices.
        # if not self.cell_centered:
        #     extents_tmp = list()
        #     remove_tmp = list()
        #     for mesh, sslc in self._subgeomslices.items():
        #         if sslc.extent not in extents_tmp:
        #             extents_tmp.append(sslc.extent)
        #         else:
        #             remove_tmp.append(mesh)
        #     for mesh in remove_tmp:
        #         del self._subgeomslices[mesh]
        #
        vals = self._subgeomslices.values()
        self.extent = Extent(min(vals, key=lambda e: e.extent.x_start).extent.x_start,
                             max(vals, key=lambda e: e.extent.x_end).extent.x_end,
                             min(vals, key=lambda e: e.extent.y_start).extent.y_start,
                             max(vals, key=lambda e: e.extent.y_end).extent.y_end,
                             min(vals, key=lambda e: e.extent.z_start).extent.z_start,
                             max(vals, key=lambda e: e.extent.z_end).extent.z_end)

        if self.extent.x_start == self.extent.x_end:
            self.orientation = 1
        elif self.extent.y_start == self.extent.y_end:
            self.orientation = 2
        elif self.extent.z_start == self.extent.z_end:
            self.orientation = 3
        else:
            self.orientation = 0

        # If lazy loading has been disabled by the user, load the data instantaneously instead
        if not settings.LAZY_LOAD:
            for _, subgeomslice in self._subgeomslices.items():
                _ = subgeomslice.data

    def get_subgeomslice(self, key: Union[int, Mesh]) -> SubGeomSlice:
        """Returns the :class:`SubGeomSlice` that cuts through the given mesh. When an int is provided
            the nth SubGeomSlice will be returned.
        """
        return self[key]

    def __getitem__(self, key: Union[int, Mesh]) -> SubGeomSlice:
        """Returns the :class:`SubGeomSlice` that cuts through the given mesh. When an int is provided
            the nth SubGeomSlice will be returned.
        """
        if type(key) == int:
            return tuple(self._subgeomslices.values())[key]
        return self._subgeomslices[key]

    def __len__(self):
        return len(self._subgeomslices)

    @property
    def subgeomslices(self) -> List[SubGeomSlice]:
        """Get a list with all SubGeomSlices.
        """
        return list(self._subgeomslices.values())

    # @property
    # def extent_dirs(self) -> Tuple[Literal['x', 'y', 'z'], Literal['x', 'y', 'z'], Literal['x', 'y', 'z']]:
    #     """The directions in which there is an extent. All three dimensions in case the geomslice is 3D.
    #     """
    #     ior = self.orientation
    #     if ior == 0:
    #         return 'x', 'y', 'z'
    #     elif ior == 1:
    #         return 'y', 'z'
    #     elif ior == 2:
    #         return 'x', 'z'
    #     else:
    #         return 'x', 'y'

    def get_nearest_timestep(self, time: float) -> int:
        """Calculates the nearest timestep for which data has been output for this geomslice.
        """
        idx = np.searchsorted(self.times, time, side="left")
        if time > 0 and (idx == len(self.times) or np.math.fabs(
                time - self.times[idx - 1]) < np.math.fabs(time - self.times[idx])):
            return idx - 1
        else:
            return idx

    # def get_nearest_index(self, dimension: Literal['x', 'y', 'z'], value: float) -> int:
    #     """Get the nearest mesh coordinate index in a specific dimension.
    #     """
    #     coords = self.coordinates[dimension]
    #     idx = np.searchsorted(coords, value, side="left")
    #     if idx > 0 and (idx == coords.size or np.math.fabs(value - coords[idx - 1]) < np.math.fabs(
    #             value - coords[idx])):
    #         return idx - 1
    #     else:
    #         return idx

    @property
    def meshes(self) -> List[Mesh]:
        """Returns a list of all meshes this geomslice cuts through.
        """
        return list(self._subgeomslices.keys())

    # @property
    # def coordinates(self) -> Dict[Literal['x', 'y', 'z'], np.ndarray]:
    #     """Returns a dictionary containing a numpy ndarray with coordinates for each dimension.
    #         For cell-centered geomslices, the coordinates are adjusted to represent cell-centered coordinates.
    #     """
    #     coords = {'x': set(), 'y': set(), 'z': set()}
    #     for dim in ('x', 'y', 'z'):
    #         for mesh in self._subgeomslices.keys():
    #             co = mesh.coordinates[dim].copy()
    #             # In case the geomslice is cell-centered, we will shift the coordinates by half a cell
    #             # and remove the last coordinate
    #             if self.cell_centered:
    #                 co = co[:-1]
    #                 co -= (co[1] - co[0]) / 2
    #             coords[dim].update(co)
    #         coords[dim] = np.array(sorted(list(coords[dim])))
    #
    #     return coords

    @property
    def vmin(self):
        return np.min(self)

    @property
    def vmax(self):
        return np.max(self)

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for subgeomslice in self._subgeomslices.values():
            subgeomslice.clear_cache()

    @property
    def vertices(self):
        n_verts = sum(subgeomslice.vertices.shape[0] for subgeomslice in self._subgeomslices.values())
        vertices = np.empty((n_verts, 3))

        counter = 0
        for subgeomslice in self._subgeomslices.values():
            size = subgeomslice.vertices.shape[0]
            vertices[:, counter:counter+size] = subgeomslice.vertices
            counter += size

        return vertices

    @property
    def faces(self):
        n_faces = sum(x.faces.shape[0] for x in self._subgeomslices.values())
        faces = np.empty((n_faces, 3))

        counter = 0
        for subgeomslice in self._subgeomslices.values():
            size = subgeomslice.faces.shape[0]
            faces[:, counter:counter + size] = subgeomslice.faces
            counter += size

        return faces

    @property
    def data(self):
        n = sum(x.data.shape[1] for x in self._subgeomslices.values())
        data = np.empty((next(iter(self._subgeomslices.values())).n_t, n))

        counter = 0
        for subgeomslice in self._subgeomslices.values():
            size = subgeomslice.data.shape[1]
            data[:, counter:counter + size] = subgeomslice.data
            counter += size

        return data

    @implements(np.min)
    def _min(self):
        return min(subgeomsclice.vmin for subgeomsclice in self._subgeomslices.values())

    @implements(np.max)
    def _max(self):
        return max(subgeomsclice.vmax for subgeomsclice in self._subgeomslices.values())

    @implements(np.mean)
    def mean(self):
        """Calculates the mean over the whole geomslice.

        :returns: The calculated mean value.
        """
        return np.mean([np.mean(subgeomsclice.data) for subgeomsclice in self._subgeomslices.values()])

    @implements(np.std)
    def std(self):
        """Calculates the standard deviation over the whole geomslice.

        :returns: The calculated standard deviation.
        """
        mean = self.mean
        sum = np.sum([np.sum(np.power(subgeomsclice.data - mean, 2)) for subgeomsclice in self._subgeomslices.values()])
        N = np.sum([subgeomsclice.data.size for subgeomsclice in self._subgeomslices.values()])
        return np.sqrt(sum / N)

    def __array__(self):
        """Method that will be called by numpy when trying to convert the object to a numpy ndarray.
        """
        raise UserWarning(
            "Slices can not be converted to numpy arrays, but they support all typical numpy"
            " operations such as np.multiply. If a 'global' array containg all subgeomslices is"
            " required, use the 'to_global' method and use the returned numpy-array explicitly.")

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Method that will be called by numpy when using a ufunction with a GeomSlice as input.

        :returns: A new geomslice on which the ufunc has been applied.
        """
        if method != "__call__":
            logging.warning(
                "The %s method has been used which is not explicitly implemented. Correctness of"
                " results is not guaranteed. If you require this feature to be implemented please"
                " submit an issue on Github where you explain your use case.", method)
        input_list = list(inputs)
        for i, inp in enumerate(inputs):
            if isinstance(inp, self.__class__):
                del input_list[i]
        if len(input_list) == 0:
            raise UserWarning(
                f"The {method} operation is not implemented for multiple geomslices as input yet. If"
                " you require this feature, please request this functionality by submitting an"
                " issue on Github.")

        new_slice = deepcopy(self)
        for subgeomslice in new_slice._subgeomslices.values():
            subgeomslice._data = ufunc(subgeomslice.data, input_list[0], **kwargs)
        return new_slice

    def __array_function__(self, func, types, args, kwargs):
        """Method that will be called by numpy when using an array function with a GeomSlice as input.

        :returns: The output of the array function.
        """
        if func not in _HANDLED_FUNCTIONS:
            return NotImplemented
            # Note: this allows subclasses that don't override __array_function__ to handle GeomSlices.
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        return _HANDLED_FUNCTIONS[func](*args, **kwargs)

    def __repr__(self):
        # if self.type == '3D':  # 3D-Slice
        #     return f"GeomSlice([3D] quantity={self.quantity}, cell_centered={self.cell_centered}, extent={self.extent})"
        # else:  # 2D-Slice
        #     return f"GeomSlice([2D] quantity={self.quantity}, cell_centered={self.cell_centered}, extent={self.extent}, " \
        #            f"extent_dirs={self.extent_dirs}, orientation={self.orientation})"
        return f"GeomSlice(quantity={self.quantity})"

# __array_function__ implementations



from typing import List, Tuple, Dict, Sequence

import numpy as np

from fdsreader.utils import Quantity, Mesh


class Particle:
    """Container to store particle data from particle simulations with FDS.

    :ivar class_name: Name of the particle class defined in the FDS input-file.
    :ivar quantities: List of all quantities for which data has been written out.
    :ivar color: Color assigned to the particle.
    :ivar n_particles: Number of existing particles for each timestep per mesh.
    :ivar lower_bounds: Dictionary with lower bounds for each timestep with quantities as keys.
    :ivar upper_bounds: Dictionary with upper bounds for each timestep with quantities as keys.
    """

    def __init__(self, class_name: str, quantities: List[Quantity], color: Tuple[float, float, float]):
        self.class_name = class_name
        self.quantities = quantities
        self.color = color
        self.n_particles: Dict[Mesh, List[int]] = dict()

        self._positions: List[np.ndarray] = list()
        self._tags: List[np.ndarray] = list()
        self._data: Dict[str, List[np.ndarray]] = {q.name: [] for q in self.quantities}
        self.times: Sequence[float] = list()

        self.lower_bounds = {q.name: [] for q in self.quantities}
        self.upper_bounds = {q.name: [] for q in self.quantities}

        self._init_callback = None

    @property
    def id(self):
        return self.class_name

    def filter_by_tag(self, tag: int):
        """Filter all particles by a single one with the specified tag.
        """
        data = self.data
        tags = self.tags
        positions = self.positions
        part = Particle(self.class_name, self.quantities, self.color)
        part._tags = tag

        part._data = {quantity: list() for quantity in data.keys()}
        part._positions = list()
        part.times = list()

        for t, tags in enumerate(tags):
            if tag in tags:
                idx = np.where(tags == tag)[0]

                for quantity in data.keys():
                    part._data[quantity].append(data[quantity][t][idx][0])
                part._positions.append(positions[t][idx][0])
                part.times.append(self.times[t])

        part.lower_bounds = dict()
        part.upper_bounds = dict()
        for q in self.quantities:
            if len(part._positions) != 0:
                part.lower_bounds[q.name] = np.min(part._data[q.name])
                part.upper_bounds[q.name] = np.max(part._data[q.name])
            else:
                part.lower_bounds[q.name] = 0
                part.upper_bounds[q.name] = 0

        return part

    @property
    def data(self) -> Dict[str, List[np.ndarray]]:
        """Dictionary with quantities as keys and a list with a numpy array for each timestep which
            contains data for each particle in that timestep.
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._data

    @property
    def tags(self) -> List[np.ndarray]:
        """List with a numpy array for each timestep which contains a tag for each particle in that
            timestep.
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._tags

    @property
    def positions(self) -> List[np.ndarray]:
        """List with a numpy array for each timestep which contains the position of each particle in
            that timestep.
        """
        if len(self._positions) == 0 and len(self._tags) == 0:
            self._init_callback()
        return self._positions

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if len(self._positions) != 0:
            del self._positions
            self._positions = list()
            del self._tags
            self._tags = list()
            del self._data
            self._data = {q.name: [] for q in self.quantities}

    def __repr__(self):
        return f"Particle(name={self.class_name}, quantities={self.quantities})"



import logging
import os
from typing import Iterable, Dict, List, Union
import numpy as np

from fdsreader.evac import Evacuation
from fdsreader.utils import Mesh
from fdsreader.utils.data import FDSDataCollection, Quantity
import fdsreader.utils.fortran_data as fdtype
from fdsreader import settings


class EvacCollection(FDSDataCollection):
    """Collection of :class:`Evacuation` objects. Next to agent-class specific data (such as trajectories) lots of
        other data such as FED-data is provided via this class.

        :ivar times: List of all time steps of the simulation.
        :ivar z_offsets: The offset in z-direction for each mesh where the evac plane lays.
        :ivar all_agents: Number of all agents per time step.
        :ivar agents_inside_mesh: Number of all agents per time step inside a specific mesh.
        :ivar number_of_deads: Number of dead agents per time step.
        :ivar fed_max: FED max per time step.
        :ivar fed_max_alive: FED max alive per time step.
        :ivar exit_counters: Exit counts per time step.
        :ivar target_exit_counters: Target exit counts per time step.
        :ivar door_counters: Door counts per time step.
        :ivar target_door_counters: Target door counts per time step.
    """

    def __init__(self, evacs: Iterable[Evacuation], base_path: str):
        super().__init__(evacs)

        self._file_paths: Dict[Mesh, str] = dict()
        self.z_offsets: Dict[Mesh, float] = dict()
        self._base_path = base_path

        self._load_csv_data()

        for evac in evacs:
            evac.times = self.times

        for evac in self:
            if settings.LAZY_LOAD:
                evac._init_callback = self._load_prt_data
            else:
                self._load_xyz_data()
                self._load_eff_data()
                self._load_prt_data()

    @property
    def quantities(self) -> List[Quantity]:
        """Gives a list of all quantities that are written out for some (or sometimes all) human classes.
        """
        qs = set()
        for evac in self:
            for q in evac.quantities:
                qs.add(q)
        return list(qs)

    def _load_csv_data(self):
        file_path = self._base_path + ".csv"
        if not os.path.exists(file_path):
            return
        with open(file_path, 'r') as infile:
            units = [unit.replace('"', '').replace('\n', '').strip() for unit in infile.readline().split(',')]
            names = [name.replace('"', '').replace('\n', '').strip() for name in infile.readline().split(',')]
            values = np.genfromtxt(infile, delimiter=',', dtype=np.float32, autostrip=True)
            dtypes = [int] * len(names)
            dtypes[0] = float
            dtypes[-2:] = (float, float)

            size = values.shape[0]
            data = {names[i]: np.empty((size,), dtype=dtypes[i]) for i in range(len(names))}
            for k, arr in enumerate(data.values()):
                for i in range(size):
                    arr[i] = values[i][k]

            self.times = data["EVAC_Time"]
            self.all_agents = data["AllAgents"]
            self.agents_inside_mesh = {names[i]: data[names[i]] for i in range(len(names)) if
                                       units[i] == "AgentsInsideMesh"}
            self.number_of_deads = data["Number_of_Deads"]
            self.fed_max = data["FED_max"]
            self.fed_max_alive = data["FED_max_alive"]

            self.exit_counters = {names[i]: data[names[i]] for i in range(len(names)) if units[i] == "ExitCounter"}
            self.target_exit_counters = {names[i]: data[names[i]] for i in range(len(names)) if
                                         units[i] == "TargetExitCounter"}

            self.door_counters = {names[i]: data[names[i]] for i in range(len(names)) if units[i] == "DoorCounter"}
            self.target_door_counters = {names[i]: data[names[i]] for i in range(len(names)) if
                                         units[i] == "TargetDoorCounter"}

    @property
    def xyz(self) -> List[np.ndarray]:
        """List of xyz-data for each mesh.
        """
        if not hasattr(self, "_xyz"):
            self._load_xyz_data()
        return self._xyz

    def _load_xyz_data(self):
        file_path = self._base_path + ".xyz"

        if not os.path.exists(file_path):
            self._xyz = np.array([])
            return

        with open(file_path, 'rb') as infile:
            dtype_meta = fdtype.new((('i', 6),))
            dtype_grid_meta = fdtype.new((('i', 4),))
            dtype_grid_data = fdtype.new((('f', 3),))

            file_format = fdtype.read(infile, fdtype.INT, 1)[0][0][0]
            if file_format != -4:
                logging.warning("The evac xyz file (" + file_path + ") was written in an unsupported file format. "
                                                                    "Please submit an issue on Github!")

            meta = fdtype.read(infile, dtype_meta, 1)[0][0]
            n_grids = meta[0]
            n_corrs = meta[2]
            n_devc = fdtype.read(infile, fdtype.INT, 1)[0][0][0]

            n_i = list()
            n_j = list()
            for g in range(n_grids):
                n_i_g, n_j_g, _, _ = fdtype.read(infile, dtype_grid_meta, 1)[0][0]
                n_i.append(n_i_g)
                n_j.append(n_j_g)
                for i in range(n_i_g):
                    for j in range(n_j_g):
                        _ = fdtype.read(infile, dtype_grid_data, 1)

            # Reset pointer (after file header)
            infile.seek(fdtype.INT.itemsize * 2 + dtype_meta.itemsize)

            self._xyz = [np.empty((n_i[g], n_j[g], 3)) for g in range(n_grids)]

            for g in range(n_grids):
                infile.seek(dtype_grid_meta.itemsize, 1)
                for i in range(n_i[g]):
                    for j in range(n_j[g]):
                        self._xyz[g][i, j] = fdtype.read(infile, dtype_grid_data, 1)[0][0]

        if os.path.exists(self._base_path + ".fed"):
            self._load_fed_data(n_corrs)

    @property
    def fed_grid(self) -> Dict[str, List[np.ndarray]]:
        """
        """
        if not hasattr(self, "_fed_grid"):
            self._load_xyz_data()
        return self._fed_grid

    @property
    def fed_corr(self) -> Dict[str, List[np.ndarray]]:
        """
        """
        if not hasattr(self, "_fed_corr"):
            self._load_xyz_data()
        return self._fed_corr

    @property
    def devc(self) -> Dict[str, Union[List[np.ndarray], np.ndarray]]:
        """
        """
        if not hasattr(self, "_devc"):
            self._load_xyz_data()
        return self._devc

    @property
    def fed_times(self) -> np.ndarray:
        """
        """
        if not hasattr(self, "_fed_times"):
            self._load_xyz_data()
        return self._fed_times

    def _load_fed_data(self, n_corrs: int):
        file_path = self._base_path + ".fed"

        if not os.path.exists(file_path):
            self._fed_grid = dict(co_co2_o2=[], soot_dens=[], tmp_g=[], radflux=[])
            self._fed_corr = dict(co_co2_o2=[], soot_dens=[], tmp_g=[], radflux=[])
            self._devc = dict(current=[], prior=[], t_change=[], times=np.array([]))
            self._fed_times = np.array([])
            return

        with open(file_path, 'rb') as infile:
            dtype_meta = fdtype.new((('i', 6),))
            dtype_time = fdtype.new((('f', 2),))
            dtype_grid_meta = fdtype.new((('i', 4),))
            dtype_corr = fdtype.new((('f', 8),))
            dtype_devs_meta = fdtype.new((('i', 2),))
            dtype_devs_data = fdtype.new((('i', 1), ('f', 1), ('i', 2), ('f', 1)))

            # File header
            file_format = fdtype.read(infile, fdtype.INT, 1)[0][0][0]
            if file_format != -4:
                logging.warning("The evac fed file (" + file_path + ") was written in an unsupported file format. "
                                                                    "Please submit an issue on Github!")

            n_grids = fdtype.read(infile, dtype_meta, 1)[0][0][0]
            n_devc = fdtype.read(infile, fdtype.INT, 1)[0][0][0]

            # Read gridsize from first timestep header
            infile.seek(dtype_time.itemsize, 1)

            n_i = list()
            n_j = list()
            n = list()
            dtype_grid_data = list()
            for g in range(n_grids):
                n_i_g, n_j_g, _, n_g = fdtype.read(infile, dtype_grid_meta, 1)[0][0]
                n_i.append(n_i_g)
                n_j.append(n_j_g)
                n.append(n_g)
                dtype_grid_data.append(fdtype.new((('f', n_g),)))
                for i in range(n_i_g):
                    for j in range(n_j_g):
                        _ = fdtype.read(infile, dtype_grid_data[-1], 1)

            # Reset pointer (after file header)
            infile.seek(dtype_meta.itemsize + fdtype.INT.itemsize * 2)

            n_t = (os.stat(file_path).st_size - (fdtype.INT.itemsize * 2 + dtype_meta.itemsize)) // (
                        dtype_time.itemsize + sum(
                    (dtype_grid_meta.itemsize + n_i[g] * n_j[g] * dtype_grid_data[g].itemsize) for g in
                    range(n_grids)) + n_corrs * dtype_corr.itemsize + fdtype.FLOAT.itemsize + n_devc * (
                                    dtype_devs_meta.itemsize + dtype_devs_data.itemsize))

            times = list()
            self._fed_grid = dict(co_co2_o2=[np.empty((n_t, n_i[g], n_j[g])) for g in range(n_grids)],
                                  soot_dens=[np.empty((n_t, n_i[g], n_j[g])) for g in range(n_grids)],
                                  tmp_g=[np.empty((n_t, n_i[g], n_j[g])) for g in range(n_grids)],
                                  radflux=[np.empty((n_t, n_i[g], n_j[g])) for g in range(n_grids)])
            self._fed_corr = dict(co_co2_o2=[np.empty((n_t, 2)) for _ in range(n_corrs)],
                                  soot_dens=[np.empty((n_t, 2)) for _ in range(n_corrs)],
                                  tmp_g=[np.empty((n_t, 2)) for _ in range(n_corrs)],
                                  radflux=[np.empty((n_t, 2)) for _ in range(n_corrs)])
            self._devc = dict(i_type=[0 for _ in range(n_devc)], devc_id=[0 for _ in range(n_devc)],
                              current=[np.empty((n_t,), dtype=int) for _ in range(n_devc)],
                              prior=[np.empty((n_t,), dtype=int) for _ in range(n_devc)],
                              t_change=[np.empty((n_t,)) for _ in range(n_devc)], times=np.empty((n_t,)))

            for t in range(n_t):
                times.append(fdtype.read(infile, dtype_time, 1)[0][0])  # t, dt_save

                for g in range(n_grids):
                    infile.seek(dtype_grid_meta.itemsize, 1)
                    for i in range(n_i[g]):
                        for j in range(n_j[g]):
                            co_co2_o2, soot_dens, tmp_g, radflux = fdtype.read(infile, dtype_grid_data[g], 1)[0][0][:4]
                            self._fed_grid["co_co2_o2"][g][t, i, j] = co_co2_o2
                            self._fed_grid["soot_dens"][g][t, i, j] = soot_dens
                            self._fed_grid["tmp_g"][g][t, i, j] = tmp_g
                            self._fed_grid["radflux"][g][t, i, j] = radflux

                # Corr
                for c in range(n_corrs):
                    co_co2_o2_1, soot_dens_1, tmp_g_1, radflux_1, co_co2_o2_2, soot_dens_2, tmp_g_2, radflux_2 = \
                        fdtype.read(infile, dtype_corr, 1)[0][0][:8]
                    self._fed_corr["co_co2_o2"][c][t] = (co_co2_o2_1, co_co2_o2_2)
                    self._fed_corr["soot_dens"][c][t] = (soot_dens_1, soot_dens_2)
                    self._fed_corr["tmp_g"][c][t] = (tmp_g_1, tmp_g_2)
                    self._fed_corr["radflux"][c][t] = (radflux_1, radflux_2)

                self._devc["times"][t] = fdtype.read(infile, fdtype.FLOAT, 1)[0][0]
                for d in range(n_devc):
                    i_type, devc_id = fdtype.read(infile, dtype_devs_meta, 1)[0][0]
                    self._devc["i_type"][d] = i_type
                    self._devc["devc_id"][d] = devc_id
                    _, _, current, prior, t_change = fdtype.read(infile, dtype_devs_data, 1)[0]
                    self._devc["current"][d][t] = current[0]
                    self._devc["prior"][d][t] = prior[0]
                    self._devc["t_change"][d][t] = t_change[0]

            self._fed_times = np.array(times)

    @property
    def eff(self):
        if not hasattr(self, "_eff"):
            self._load_eff_data()
        return self._eff

    def _load_eff_data(self):
        file_path = self._base_path + ".eff"
        if not os.path.exists(file_path):
            return

        with open(file_path, 'rb') as infile:
            dtype_grid_meta = fdtype.new((('i', 3),))
            dtype_grid_data = fdtype.new((('f', 2),))  # u, v

            n_grids = fdtype.read(infile, fdtype.INT, 1)[0][0][0]

            # n_fields = (os.stat(file_path).st_size - fdtype.INT.itemsize) // (
            #         sum(dtype_grid_meta.itemsize + n_i[g] * n_j[g] * dtype_grid_data.itemsize for g in range(n_grids)))
            #
            # self._eff = np.empty((n_grids, n_fields, n_i, n_j, 2))
            # for g in range(n_grids):
            #     for f in range(n_fields):
            #         infile.seek(dtype_grid_meta.itemsize, 1)
            #         for i in range(n_i[g]):
            #             for j in range(n_j[g]):
            #                 self._eff[g, f, i, j] = fdtype.read(infile, dtype_grid_data, 1)[0][0]

            self._eff = list()
            meta = fdtype.read(infile, dtype_grid_meta, 1)
            while len(meta) > 0:
                n_i, n_j, _ = meta[0][0]
                for i in range(n_i + 2):
                    for j in range(n_j + 2):
                        self._eff.append(fdtype.read(infile, dtype_grid_data, 1)[0][0])
                meta = fdtype.read(infile, dtype_grid_meta, 1)

    def get_unfiltered_positions(self, quantity: Union[Quantity, str] = None):
        """Convenience function that combines all trajectories into a single array.

        :param quantity: Optionally a quantity can be specified to filter all human classes by those who have written
            out data for the specific quantity. Can either be a string or :class:`Quantity` object.
        """
        filtered_evacs = [evac for evac in self if evac.has_quantity(quantity)] if quantity is not None else self

        combined_positions = list()

        for t in range(len(self.times)):
            size = 0
            for evac in filtered_evacs:
                size += evac.positions[t].shape[0]

            combined_positions.append(np.empty((size, 3)))
            if size != 0:
                size = 0
                for evac in filtered_evacs:
                    pos = evac.positions[t]
                    combined_positions[t][size:size + pos.shape[0]] = pos
                    size += pos.shape[0]

        return combined_positions

    def get_unfiltered_data(self, quantity: Union[Quantity, str]):
        """Convenience function that combines data for a specific quantity of all humans into a single array.

        :param quantity: The quantity to get data for. Can either be a string or :class:`Quantity` object.
        """
        filtered_evacs = [evac for evac in self if evac.has_quantity(quantity)]

        combined_data = list()
        data = list()
        for evac in filtered_evacs:
            data.append(evac.get_data(quantity))

        for t in range(len(self.times)):
            size = 0
            for d in data:
                size += d[t].shape[0]

            combined_data.append(np.empty((size,)))
            if size != 0:
                size = 0
                for d in data:
                    combined_data[t][size:size + d[t].shape[0]] = d[t]
                    size += d[t].shape[0]

        return combined_data

    def _load_prt_data(self):
        """Function to read in all evac data for a simulation.
        """
        evacs = self
        pointer_location = {evac: [0] * len(self.times) for evac in evacs}

        for evac in evacs:
            if len(evac._positions) == 0 and len(evac._tags) == 0:
                for t in range(len(self.times)):
                    size = 0
                    for mesh in self._file_paths.keys():
                        size += evac.n_humans[mesh][t]
                    for quantity in evac.quantities:
                        evac._data[quantity.name].append(np.empty((size,), dtype=np.float32))
                    evac._positions.append(np.empty((size, 3), dtype=np.float32))
                    evac._body_angles.append(np.empty((size,), dtype=np.float32))
                    evac._semi_major_axis.append(np.empty((size,), dtype=np.float32))
                    evac._semi_minor_axis.append(np.empty((size,), dtype=np.float32))
                    evac._agent_heights.append(np.empty((size,), dtype=np.float32))
                    evac._tags.append(np.empty((size,), dtype=int))

        for mesh, file_path in self._file_paths.items():
            with open(file_path, 'rb') as infile:
                # Initial offset (ONE, fds version and number of evac classes)
                offset = 3 * fdtype.INT.itemsize
                # Number of quantities for each evac class (plus an INTEGER_ZERO)
                offset += fdtype.new((('i', 2),)).itemsize * len(evacs)
                # 30-char long name and unit information for each quantity
                offset += fdtype.new((('c', 30),)).itemsize * 2 * sum(
                    [len(evac.quantities) for evac in evacs])
                infile.seek(offset)

                for t in range(len(self.times)):
                    # Skip time value and meta data
                    infile.seek(fdtype.FLOAT.itemsize, 1)

                    # Read data for each evac class
                    for evac in evacs:
                        # Read number of evacs in each class
                        n_humans = fdtype.read(infile, fdtype.INT, 1)[0][0][0]
                        offset = pointer_location[evac][t]
                        # Read positions
                        dtype_positions = fdtype.new((('f', 7 * n_humans),))
                        pos = fdtype.read(infile, dtype_positions, 1)[0][0].reshape((n_humans, 7),
                                                                                    order='F').astype(float)
                        evac._positions[t][offset: offset + n_humans] = pos[:, :3]
                        evac._body_angles[t][offset: offset + n_humans] = pos[:, 3]
                        evac._semi_major_axis[t][offset: offset + n_humans] = pos[:, 4]
                        evac._semi_minor_axis[t][offset: offset + n_humans] = pos[:, 5]
                        evac._agent_heights[t][offset: offset + n_humans] = pos[:, 6]
                        # Read tags
                        dtype_tags = fdtype.new((('i', n_humans),))
                        evac._tags[t][offset: offset + n_humans] = fdtype.read(infile, dtype_tags, 1)[0][0]

                        # Read actual quantity values
                        if len(evac.quantities) > 0:
                            dtype_data = fdtype.new(
                                (('f', str((n_humans, len(evac.quantities)))),))
                            data_raw = fdtype.read(infile, dtype_data, 1)[0][0].reshape(
                                (n_humans, len(evac.quantities)), order='F')

                            for q, quantity in enumerate(evac.quantities):
                                evac._data[quantity.name][t][
                                offset:offset + n_humans] = data_raw[:, q].astype(float)
                        pointer_location[evac][t] += evac.n_humans[mesh][t]

    def __getitem__(self, key):
        """Get the evac data for a specific agent/human class.
        """
        if type(key) == int:
            return self._elements[key]
        for evac in self:
            if evac.class_name == key:
                return evac

    def __contains__(self, value):
        if value in self._elements:
            return True
        for evac in self:
            if evac.class_name == value:
                return True
        return False

    def __repr__(self):
        return "EvacCollection(" + super(EvacCollection, self).__repr__() + ")"



from typing import Iterable, Tuple, List
import numpy as np

from fdsreader.bndf import Obstruction
from fdsreader.utils.data import FDSDataCollection, Quantity


class ObstructionCollection(FDSDataCollection):
    """Collection of :class:`Obstruction` objects. Offers extensive functionality for filtering and
        using obstructions as well as dependent such as :class:`Boundary`.
    """

    def __init__(self, *obstructions: Iterable[Obstruction]):
        super().__init__(*obstructions)

    @property
    def quantities(self) -> List[Quantity]:
        qs = set()
        for obst in self:
            for q in obst.quantities:
                qs.add(q)
        return list(qs)

    def filter_by_boundary_data(self):
        """Filters all obstructions for which output data exists.
        """
        return ObstructionCollection(x for x in self if x.has_boundary_data)

    def get_by_id(self, obst_id: str):
        """Get the obstruction with corresponding id if it exists.
        """
        return next((obst for obst in self if obst.id == obst_id), None)

    def get_nearest_obstruction(self, point: Tuple[float, float, float]) -> Obstruction:
        """Filters the obstruction with the shortest distance to the given point.
        """
        d_min = np.finfo(float).max
        obst_min = None

        for obst in self:
            for subobst in obst:
                dx = max(subobst.extent.x_start - point[0], 0, point[0] - subobst.extent.x_end)
                dy = max(subobst.extent.y_start - point[1], 0, point[1] - subobst.extent.y_end)
                dz = max(subobst.extent.z_start - point[2], 0, point[2] - subobst.extent.z_end)
                d = np.sqrt(dx * dx + dy * dy + dz * dz)
                if d < d_min:
                    d_min = d
                    obst_min = obst

        return obst_min

    def __repr__(self):
        return "ObstructionCollection(" + super(ObstructionCollection, self).__repr__() + ")"



from typing import List

from fdsreader.bndf import Patch


def sort_patches_cartesian(patches_in: List[Patch]):
    """Returns all patches (of same orientation!) sorted in cartesian coordinates.
    """
    patches = patches_in.copy()
    if len(patches) != 0:
        patches_cart = [[patches[0]]]
        orientation = abs(patches[0].orientation)
        if orientation == 1:  # x
            patches.sort(key=lambda p: (p.extent.y_start, p.extent.z_start))
        elif orientation == 2:  # y
            patches.sort(key=lambda p: (p.extent.x_start, p.extent.z_start))
        elif orientation == 3:  # z
            patches.sort(key=lambda p: (p.extent.x_start, p.extent.y_start))

        if orientation == 1:
            for patch in patches[1:]:
                if patch.extent.y_start == patches_cart[-1][-1].extent.y_start:
                    patches_cart[-1].append(patch)
                else:
                    patches_cart.append([patch])
        else:
            for patch in patches[1:]:
                if patch.extent.x_start == patches_cart[-1][-1].extent.x_start:
                    patches_cart[-1].append(patch)
                else:
                    patches_cart.append([patch])
        return patches_cart
    return patches


import logging
import os
from copy import deepcopy
from typing import Dict
import numpy as np

from fdsreader.utils import Quantity, Mesh
from fdsreader import settings
import fdsreader.utils.fortran_data as fdtype

_HANDLED_FUNCTIONS = {np.mean: (lambda pl: pl.mean)}


def implements(np_function):
    """Decorator to register an __array_function__ implementation for Slices.
    """

    def decorator(func):
        _HANDLED_FUNCTIONS[np_function] = func
        return func

    return decorator


class SubSmoke3D:
    """Part of a smoke3d output for a single mesh.

    :ivar mesh: The mesh containing the data.
    :ivar upper_bounds: Numpy ndarray containing the maxmimum data value for each timestep.
    :ivar times: Numpy ndarray containing all time steps for which data has been written out.
    """

    def __init__(self, file_path: str, mesh: Mesh, upper_bounds: np.ndarray, times: np.ndarray):
        self.mesh = mesh
        self.upper_bounds = upper_bounds
        self.times = times

        self._file_path = file_path  # Path to the binary data file

    @property
    def data(self) -> np.ndarray:
        """Method to lazy load the Smoke3D data of a single mesh.
        """
        if not hasattr(self, "_data"):
            with open(self._file_path, 'rb') as infile:
                dtype_header = fdtype.new((('i', 8),))
                dtype_nchars = fdtype.new((('i', 2),))

                header = fdtype.read(infile, dtype_header, 1)[0][0]
                nx, ny, nz = int(header[3]), int(header[5]), int(header[7])
                data_shape = (nx + 1, ny + 1, nz + 1)

                self._data = np.empty((self.times.size,) + data_shape)

                for t in range(self.times.size):
                    fdtype.read(infile, fdtype.FLOAT, 1)  # Skip time value
                    nchars_out = int(fdtype.read(infile, dtype_nchars, 1)[0][0][1])

                    if nchars_out > 0:
                        dtype_data = fdtype.new((('u', nchars_out),))

                        rle_data = fdtype.read(infile, dtype_data, 1)[0][0]
                        decoded_data = np.empty(((nx + 1) * (ny + 1) * (nz + 1)))

                        # Decode run-length-encoded data (see "RLE" subroutine in smvv.f90)
                        i = 0
                        mark = np.uint8(255)
                        out_pos = 0
                        while i < nchars_out:
                            if rle_data[i] == mark:
                                value = rle_data[i+1]
                                repeats = rle_data[i+2]
                                i += 3
                            else:
                                value = rle_data[i]
                                repeats = 1
                                i += 1
                            decoded_data[out_pos:out_pos+repeats] = value
                            out_pos += repeats

                        self._data[t, :, :, :] = decoded_data.reshape(data_shape, order='F')

        return self._data

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if hasattr(self, "_data"):
            del self._data


class Smoke3D(np.lib.mixins.NDArrayOperatorsMixin):
    """Smoke3D file data container including metadata. Consists of multiple subsmokes, one for each
        mesh.

    :ivar times: Numpy ndarray containing all time steps for which data has been written out.
    :ivar quantity: :class:`Quantity` object containing information about the recorded quantity and its unit.
    """

    def __init__(self, root_path: str, times: np.ndarray, quantity: Quantity):
        self._root_path = root_path
        self.times = times
        self.quantity = quantity

        # List of all subsmokes this Smoke3D consists of (one per mesh).
        self._subsmokes: Dict[Mesh, SubSmoke3D] = dict()

    def _add_subsmoke(self, filename: str, mesh: Mesh, upper_bounds: np.ndarray) -> SubSmoke3D:
        subsmoke = SubSmoke3D(os.path.join(self._root_path, filename), mesh, upper_bounds, self.times)
        self._subsmokes[mesh] = subsmoke

        # If lazy loading has been disabled by the user, load the data instantaneously instead
        if not settings.LAZY_LOAD:
            _ = subsmoke.data

        return subsmoke

    def __getitem__(self, mesh: Mesh):
        """Returns the :class:`SubSmoke` that contains data for the given mesh.
        """
        return self.get_subsmoke(mesh)

    def get_subsmoke(self, mesh: Mesh):
        """Returns the :class:`SubSmoke` that contains data for the given mesh.
        """
        return self._subsmokes[mesh]

    @property
    def vmax(self):
        """Maximum value of all data at any time.
        """
        curr_max = max(np.max(subsmoke3d.upper_bounds) for subsmoke3d in self._subsmokes.values())
        if curr_max == np.float32(-1e33):
            return max(np.max(subsmoke3d.data) for subsmoke3d in self._subsmokes.values())
        return curr_max

    @implements(np.mean)
    def mean(self) -> np.ndarray:
        """Calculates the mean value of all Smoke3D data for this quantity.
        """
        return np.sum([np.mean(subsmoke.data) for subsmoke in self._subsmokes.values()]) / len(self._subsmokes)

    @implements(np.std)
    def std(self) -> np.ndarray:
        """Calculates the standard deviation of all Smoke3D data for this quantity.
        """
        mean = self.mean
        sum = np.sum([np.sum(np.power(subsmoke.data - mean, 2)) for subsmoke in self._subsmokes.values()])
        N = np.sum([subsmoke.data.size for subsmoke in self._subsmokes.values()])
        return np.sqrt(sum / N)

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for subsmoke in self._subsmokes.values():
            subsmoke.clear_cache()

    def __array__(self):
        """Method that will be called by numpy when trying to convert the object to a numpy ndarray.
        """
        raise UserWarning(
            "Smoke3Ds can not be converted to numpy arrays, but they support all typical numpy"
            " operations such as np.multiply. If a 'global' array containg all subsmokes is"
            " required, please request this functionality by submitting an issue on Github.")

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Method that will be called by numpy when using a ufunction with a Smoke3D as input.

        :returns: A new smoke3d on which the ufunc has been applied.
        """
        if method != "__call__":
            logging.warning(
                "The %s method has been used which is not explicitly implemented. Correctness of"
                " results is not guaranteed. If you require this feature to be implemented please"
                " submit an issue on Github where you explain your use case.", method)
        input_list = list(inputs)
        for i, inp in enumerate(inputs):
            if isinstance(inp, self.__class__):
                del input_list[i]

        new_smoke3d = deepcopy(self)
        for subsmoke in self._subsmokes.values():
            subsmoke._data = ufunc(subsmoke.data, input_list[0], **kwargs)
        return new_smoke3d

    def __array_function__(self, func, types, args, kwargs):
        """Method that will be called by numpy when using an array function with a Slice as input.

        :returns: The output of the array function.
        """
        if func not in _HANDLED_FUNCTIONS:
            return NotImplemented
            # Note: this allows subclasses that don't override __array_function__ to handle Smoke3Ds.
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        return _HANDLED_FUNCTIONS[func](*args, **kwargs)

# __array_function__ implementations
# ...



from typing import Iterable, List, Union

from fdsreader.smoke3d import Smoke3D
from fdsreader.utils.data import FDSDataCollection, Quantity


class Smoke3DCollection(FDSDataCollection):
    """Collection of :class:`Smoke3D` objects. Offers extensive functionality for filtering and
        using Smoke3Ds as well as its subclasses such as :class:`SubSmoke3D`.
    """

    def __init__(self, *smoke3ds: Iterable[Smoke3D]):
        super().__init__(*smoke3ds)

    @property
    def quantities(self) -> List[Quantity]:
        return [smoke3d.name for smoke3d in self]

    def get_by_quantity(self, quantity: Union[Quantity, str]):
        """Gets the :class:`Smoke3D`s with a specific quantity.
        """
        if type(quantity) == Quantity:
            quantity = quantity.name
        return next(x for x in self if
                    x.quantity.name.lower() == quantity.lower() or x.quantity.short_name.lower() == quantity.lower())

    def __repr__(self):
        return "Smoke3DCollection(" + super(Smoke3DCollection, self).__repr__() + ")"



import os
from typing import List, Dict, Tuple, Union, Sequence
from typing_extensions import Literal
import numpy as np

from fdsreader.utils import Surface, Extent, Quantity, Dimension
import fdsreader.utils.fortran_data as fdtype
from fdsreader import settings


class Patch:
    """Container for the actual data which is stored as rectangular plane with specific orientation
        and extent.

    :ivar dimension: :class:`Dimension` object containing information about steps in each dimension.
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    :ivar orientation: The direction the patch is facing (x={-1;1}, y={-2;2}, z={-3;3}).
    :ivar cell_centered: Indicates whether centered positioning for data is used.
    :ivar n_t: Total number of time steps for which output data has been written.
    """

    def __init__(self, file_path: str, dimension: Dimension, extent: Extent, orientation: int, cell_centered: bool,
                 patch_offset: int, initial_offset: int, n_t: int):
        self.file_path = file_path
        self.dimension = dimension
        self.extent = extent
        self.orientation = orientation
        self.cell_centered = cell_centered
        self._patch_offset = patch_offset
        self._initial_offset = initial_offset
        self._time_offset = -1
        self.n_t = n_t

    @property
    def shape(self) -> Tuple:
        """Convenience function to calculate the shape of the array containing data for this patch.
        """
        return self.dimension.shape(self.cell_centered)

    @property
    def size(self) -> int:
        """Convenience function to calculate the number of data points in the array for this patch.
        """
        return self.dimension.size(self.cell_centered)

    def _post_init(self, time_offset: int):
        """Fully initialize the patch as soon as the number of timesteps is known.
        """
        self._time_offset = time_offset

    @property
    def data(self):
        """Method to load the quantity data for a single patch.
        """
        if not hasattr(self, "_data"):
            dtype_data = fdtype.new((('f', self.dimension.size(cell_centered=False)),))

            if self.n_t == -1:
                self.n_t = (os.stat(self.file_path).st_size - self._initial_offset) // self._time_offset

            self._data = np.empty((self.n_t,) + self.shape)
            with open(self.file_path, 'rb') as infile:
                for t in range(self.n_t):
                    infile.seek(self._initial_offset + self._patch_offset + t * self._time_offset)
                    data = np.fromfile(infile, dtype_data, 1)[0][1].reshape(
                        self.dimension.shape(cell_centered=False), order='F')
                    if self.cell_centered:
                        self._data[t, :] = data[:-1, :-1]
                    else:
                        self._data[t, :] = data
        return self._data

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if hasattr(self, "_data"):
            del self._data

    def __repr__(self, *args, **kwargs):
        return f"Patch(shape={self.shape}, orientation={self.orientation}, extent={self.extent})"


class Boundary:
    """Container for boundary data specific to one quantity.

    :ivar quantity: Quantity object containing information about the quantity calculated for this
        :class:`Obstruction` with the corresponding short_name and unit.
    :ivar times: Numpy array containing all times for which data has been recorded.
    :ivar cell_centered: Indicates whether centered positioning for data is used.
    :ivar lower_bounds: Dictionary with lower bounds for each timestep with meshes as keys.
    :ivar upper_bounds: Dictionary with upper bounds for each timestep with meshes as keys.
    :ivar n_t: Total number of time steps for which output data has been written.
    """

    def __init__(self, quantity: Quantity, cell_centered: bool, times: np.ndarray, n_t: int, patches: List[Patch],
                 lower_bounds: np.ndarray, upper_bounds: np.ndarray):
        self.quantity = quantity
        self.cell_centered = cell_centered
        self.patches = patches
        self.times = times
        self.n_t = n_t
        self.lower_bounds = lower_bounds
        self.upper_bounds = upper_bounds

    @property
    def data(self) -> Dict[int, Patch]:
        """The :class:`Patch` in each direction (-3=-z, -2=-y, -1=-x, 1=x, 2=y, 3=y).
        """
        return {p.orientation: p for p in self.patches}

    def vmin(self, orientation: Literal[-3, -2, -1, 0, 1, 2, 3] = 0) -> float:
        """Minimum value of all patches at any time.

        :param orientation: Optionally filter by patches with a specific orientation.
        """
        if orientation == 0:
            curr_min = np.min(self.lower_bounds)
            if curr_min == 0.0:
                return min(np.min(p.data) for p in self.patches)
            return curr_min
        else:
            return np.min(self.data[orientation].data)

    def vmax(self, orientation: Literal[-3, -2, -1, 0, 1, 2, 3] = 0) -> float:
        """Maximum value of all patches at any time.

        :param orientation: Optionally filter by patches with a specific orientation.
        """
        if orientation == 0:
            curr_max = np.max(self.upper_bounds)
            if curr_max == np.float32(-1e33):
                return max(np.max(p.data) for p in self.patches)
            return curr_max
        else:
            return np.max(self.data[orientation].data)

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for p in self.patches:
            p.clear_cache()

    def __repr__(self):
        return f"Boundary(Quantity={self.quantity}, Patches={len(self.patches)})"


class SubObstruction:
    """An :class:`Obstruction` consists of 1 or more SubObstructions which can be hidden at specific points in time.

    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    :ivar bound_indices: Indices used to define obstruction bounds in terms of mesh locations.
    :ivar side_surfaces: Tuple of six surfaces for each side of the cuboid.
    :ivar hide_times: List with points in time from when on the SubObstruction will be hidden.
    :ivar show_times: List with points in time from when on the SubObstruction will be shown.
    """

    def __init__(self, side_surfaces: Tuple[Surface, ...], bound_indices: Tuple[int, int, int, int, int, int],
                 extent: Extent):
        self.extent = extent
        self.side_surfaces = side_surfaces
        self.bound_indices = {'x': (bound_indices[0], bound_indices[1]),
                              'y': (bound_indices[2], bound_indices[3]),
                              'z': (bound_indices[4], bound_indices[5])}

        self._boundary_data: Dict[int, Boundary] = dict()

        self.hide_times = list()
        self.show_times = list()

    def _add_patches(self, bid: int, cell_centered: bool, quantity: str, short_name: str, unit: str, patches: List[Patch],
                     times: np.ndarray, n_t: int, lower_bounds: np.ndarray, upper_bounds: np.ndarray):
        if bid not in self._boundary_data:
            self._boundary_data[bid] = Boundary(Quantity(quantity, short_name, unit), cell_centered, times, n_t, patches,
                                                lower_bounds, upper_bounds)

        if not settings.LAZY_LOAD:
            _ = self._boundary_data[bid].data

    def get_data(self, quantity: Union[str, Quantity]):
        if type(quantity) == Quantity:
            quantity = quantity.name
        return next(b for b in self._boundary_data.values() if
                    b.quantity.name.lower() == quantity.lower() or b.quantity.short_name.lower() == quantity.lower())

    def __getitem__(self, item):
        if type(item) == int:
            return self._boundary_data[item]
        return self.get_data(item)

    def _hide(self, time: float):
        self.hide_times.append(time)
        self.hide_times.sort()

    def _show(self, time: float):
        self.show_times.append(time)
        self.show_times.sort()

    def visible_times(self, times: Sequence[float]) -> np.ndarray:
        """Returns an ndarray containing all time steps when there is data available for the SubObstruction. Will return an
            empty list when no data is output at all.

        :param times: All timesteps of the simulation.
        """
        ret = list()

        hidden = False
        for time in times:
            if time in self.show_times:
                hidden = False
            if time in self.hide_times:
                hidden = True
            if not hidden:
                ret.append(time)

        return np.array(ret)

    def vmin(self, quantity: Union[str, Quantity], orientation: Literal[-3, -2, -1, 0, 1, 2, 3] = 0) -> float:
        """Minimum value of all patches at any time for a specific quantity.

        :param orientation: Optionally filter by patches with a specific orientation.
        """
        return self.get_data(quantity).vmin(orientation)

    def vmax(self, quantity: Union[str, Quantity], orientation: Literal[-3, -2, -1, 0, 1, 2, 3] = 0) -> float:
        """Maximum value of all patches at any time for a specific quantity.

        :param orientation: Optionally filter by patches with a specific orientation.
        """
        return self.get_data(quantity).vmax(orientation)

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for bndf in self._boundary_data.values():
            bndf.clear_cache()

    def __repr__(self):
        return f"SubObstruction(Extent={self.extent})"


class Obstruction:
    """A box-shaped obstruction with specific surfaces (materials) on each side.

    :ivar id: ID of the obstruction.
    :ivar color_index: Type of coloring used to color obstruction.
     \n-1 - default color
     \n-2 - invisible
     \n-3 - use red, green, blue and alpha (rgba attribute)
     \nn>0 - use n’th color table entry
    :ivar block_type: Defines how the obstruction is drawn.
     \n-1 - use surface to obtain blocktype
     \n0 - regular block
     \n2 - outline
    :ivar texture_origin: Origin position of the texture provided by the surface. When the texture
        does have a pattern, for example windows or bricks, the texture_origin specifies where the
        pattern should begin.
    :ivar rgba: Optional color of the obstruction in form of a 4-element tuple
        (ranging from 0.0 to 1.0).
    """

    def __init__(self, oid: int, color_index: int, block_type: int, texture_origin: Tuple[float, float, float],
                 rgba: Union[Tuple[()], Tuple[float, float, float, float]] = ()):
        self.id = oid
        self.color_index = color_index
        self.block_type = block_type
        self.texture_origin = texture_origin
        if len(rgba) != 0:
            self.rgba = rgba

        self._subobstructions: List[SubObstruction] = list()

    @property
    def bounding_box(self) -> Extent:
        """:class:`Extent` object representing the bounding box around the Obstruction.
        """
        extents = [sub.extent for sub in self._subobstructions]

        return Extent(min(extents, key=lambda e: e.x_start).x_start, max(extents, key=lambda e: e.x_end).x_end,
                      min(extents, key=lambda e: e.y_start).y_start, max(extents, key=lambda e: e.y_end).y_end,
                      min(extents, key=lambda e: e.z_start).z_start, max(extents, key=lambda e: e.z_end).z_end)

    @property
    def quantities(self) -> List[Quantity]:
        """Get a list of all quantities for which boundary data exists.
        """
        return [b.quantity for b in self._subobstructions[0]._boundary_data.values()]

    def filter_by_orientation(self, orientation: Literal[-3, -2, -1, 0, 1, 2, 3] = 0) -> List[SubObstruction]:
        """Filter all SubObstructions by a specific orientation. All returned SubObstructions will contain boundary data
            in the specified orientation.
        """
        return [subobst for subobst in self._subobstructions if
                orientation in next(iter(subobst._boundary_data.values())).data.keys()]

    def get_boundary_data(self, quantity: Union[Quantity, str],
                          orientation: Literal[-3, -2, -1, 0, 1, 2, 3] = 0) -> List[Boundary]:
        """Gets the boundary data for a specific quantity of all SubObstructions.

        :param quantity: The quantity to filter by.
        :param orientation: Optionally filter by a specific orientation as well (-3=-z, -2=-y, -1=-x, 1=x, 2=y, 3=z).
            A value of 0 indicates to no filter.
        """
        if type(quantity) == Quantity:
            quantity = quantity.name

        ret = [subobst.get_data(quantity) for subobst in self._subobstructions]
        if orientation == 0:
            return ret
        return [bndf for bndf in ret if orientation in bndf.data.keys()]

    @property
    def has_boundary_data(self):
        """Whether boundary data has been output in the simulation.
        """
        return len(self._subobstructions[0]._boundary_data) != 0

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for s in self._subobstructions:
            s.clear_cache()

    def vmin(self, quantity: Union[str, Quantity], orientation: Literal[-3, -2, -1, 0, 1, 2, 3] = 0) -> float:
        """Minimum value of all patches at any time for a specific quantity.

        :param orientation: Optionally filter by patches with a specific orientation.
        """
        return min(s.vmin(quantity, orientation) for s in self._subobstructions)

    def vmax(self, quantity: Union[str, Quantity], orientation: Literal[-3, -2, -1, 0, 1, 2, 3] = 0) -> float:
        """Maximum value of all patches at any time for a specific quantity.

        :param orientation: Optionally filter by patches with a specific orientation.
        """
        return max(s.vmax(quantity, orientation) for s in self._subobstructions)

    def __getitem__(self, index):
        """Gets the nth :class:`SubObstruction`.
        """
        return self._subobstructions[index]

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self, *args, **kwargs):
        return f"Obstruction(id={self.id}, Bounding-Box={self.bounding_box}, SubObstructions={len(self._subobstructions)}" + (
            f", Quantities={[q.short_name for q in self.quantities]}" if self.has_boundary_data else "") + ")"
